""" This module provides the infrastructure for loopy preambles that have
a complex requirement structure. This includes:
* preambles triggering the creation of other preambles
* a caching mechanism to avoid duplicated preambles where harmful
"""


class UFL2LoopyDataCache(dict):
    """ The cache data structure

    The data is stored as key value pairs of the following form:
    (function, cache_tuple) -> (content, priority_tag, preamble)

    The parameters are:
    function : the function that generates the preamble code snippet
    cache_tuple : A frozen (sub)set of arguments to the function.
                  The map from function arguments to cache tuple controls
                  the amount of caching.
    content : The content to store. No assumptions made.
    priority_tag : Will later decide the ordering of the preambles.
    preamble : A bool whether this cache entry does generate a loopy preamble.
               This is usually not the case, when you generate something
               like symbol names. Those are inserted into other snippets,
               but not on their own right.
    """
    def __init__(self):
        self.counter = 0

    def register(self, cachekey, content):
        self[cachekey] = (content, self.counter)
        self.counter = self.counter + 1

    def extract_preambles(self):
        return [(p[1], p[0]) for p in self.values() if p[0].generate_preambel]


# have one such cache on the module level. It is easier than handing around an instance of it.
_cache = UFL2LoopyDataCache()


def freeze(data):
    """ A function that deterministically generates an
    immutable item from the given data. Used for cache key generation.
    """
    from collections import Hashable, Iterable, MutableMapping

    # Strings are iterable and hashable, but they do not pose any problems.
    if isinstance(data, str):
        return data

    import ufl.classes
    if isinstance(data, ufl.classes.Expr):
        return data

    # Check if the given data is already hashable
    if isinstance(data, Hashable):
        if isinstance(data, Iterable):
            return type(data)(freeze(i) for i in data)
        return data

    # convert standard mutable containers
    if isinstance(data, MutableMapping):
        return tuple((freeze(k), freeze(v)) for k,v in data.iteritems())
    if isinstance(data, Iterable):
        return tuple(freeze(i) for i in data)

    # we don't know how to handle this object, so we give up
    raise TypeError('Cannot freeze non-hashable object {} of type {}'.format(data, type(data)))


class _RegisteredFunction(object):
    """ The data structure for a function that accesses UFL2LoopyDataCache """
    def __init__(self, func,
                 cache_key_generator=lambda *a : freeze(a),
                 generate_preamble=False,
                 return_content=False
                ):
        self.func = func
        self.cache_key_generator = cache_key_generator
        self.generate_preamble = generate_preamble
        self.return_content = return_content

    def __call__(self, *args):
        cache_key = (self, self.cache_key_generator(*args))
        if cache_key in _cache:
            if self.return_content:
                return _cache[cache_key][0]
        else:
            content = self.func(*args)
            _cache.register(cache_key, content)
            if self.return_content:
                return content


def _dune_decorator_factory(**factory_kwargs):
    """ A function decorator factory

    Generates a function decorator, that turns a given function into
    a function that stores its result in the UFL2LoopyDataCache.

    The generated decorator may be used with or without keyword arguments.

    The keyword arguments given to this factory override any keywords
    given to the decorator. This allows the definition of specialized
    decorators for different purposes.

    Possible keyword arguments:
    ---------------------------
    cache_key_generator : function
        A function that maps the arguments to the function to an immutable cache key.
        Defaults to generate_cache_tuple.
    generate_preambel : bool
        Whether a code snippet should be generated.
        This is usually not the case, when you generate something
        like symbol names. Those are inserted into other snippets,
        but not on their own right.
    return_content : bool
        Whether the content of the cache should be returned on subsequent function calls.
        If the cache is only used for requirements tracking, this is not needed.
    """
    def _dec(*args, **kwargs):
        # Modify the kwargs according to the factorys kwargs
        for k in factory_kwargs:
            kwargs[k] = factory_kwargs[k]
        if args:
            assert len(args) == 1
            return _RegisteredFunction(args[0], **kwargs)
        else:
            return lambda f: _RegisteredFunction(f, **kwargs)

    return _dec


# A decorator for a dune preambel
dune_preambel = _dune_decorator_factory(generate_preamble=True)

# A decorator for a dune symbol name
dune_symbol = _dune_decorator_factory(return_content=True)
